# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import json

from django import http


from django.db.models import F

from dingos.models import Identifier, InfoObject2Fact, InfoObject
from dingos.filter import InfoObjectFilter, FactTermValueFilter, IdSearchFilter

from dingos import DINGOS_TEMPLATE_FAMILY

from view_classes import BasicFilterView, BasicDetailView, BasicTemplateView



class InfoObjectList(BasicFilterView):

    template_name = 'dingos/%s/lists/InfoObjectList.html' % DINGOS_TEMPLATE_FAMILY

    breadcrumbs = (('Dingo',None),
                   ('List',None),
                   ('InfoObject',None))

    paginate_by = 15

    filterset_class= InfoObjectFilter

    title = 'List of Info Objects (generic filter)'

    queryset = InfoObject.objects.exclude(latest_of=None).prefetch_related(
        'iobject_type',
        'iobject_type__iobject_family',
        'iobject_family',
        'identifier__namespace',
        'iobject_family_revision',
        'identifier').select_related().distinct().order_by('-latest_of__pk')

class InfoObjectList_Id_filtered(BasicFilterView):

    template_name = 'dingos/%s/lists/InfoObjectList.html' % DINGOS_TEMPLATE_FAMILY

    breadcrumbs = (('Dingo',None),
                   ('List',None),
                   ('InfoObject',None))

    paginate_by = 15

    filterset_class= IdSearchFilter

    title = 'ID-based filtering'

    queryset = InfoObject.objects.exclude(latest_of=None).prefetch_related(
        'iobject_type',
        'iobject_family',
        'iobject_family_revision',
        'identifier').select_related().distinct().order_by('-latest_of__pk')

class InfoObjectsEmbedded(BasicFilterView):
    template_name = 'dingos/%s/lists/InfoObjectEmbedded.html' % DINGOS_TEMPLATE_FAMILY

    breadcrumbs = (('Dingo',None),
                   ('List',None),
                   ('InfoObject',None))

    paginate_by = 15

    filterset_class = InfoObjectFilter

    @property
    def title(self):
        return 'Objects that embed object "%s" (id %s)' % (self.iobject.name,
                                                           self.iobject.identifier)
    @property
    def iobject(self):
        return InfoObject.objects.get(pk=int(self.kwargs['pk']))

    def get_queryset(self):

        queryset = InfoObject2Fact.objects.exclude(iobject__latest_of=None). \
            filter(fact__value_iobject_id__id=self.iobject.identifier.id). \
            filter(iobject__timestamp=F('iobject__identifier__latest__timestamp')). \
            order_by('-iobject__timestamp')
        return queryset

    def get_context_data(self, **kwargs):
        context = super(InfoObjectsEmbedded, self).get_context_data(**kwargs)
        context['iobject'] = self.iobject
        return context


class SimpleFactSearch(BasicFilterView):
    template_name = 'dingos/%s/searches/SimpleFactSearch.html' % DINGOS_TEMPLATE_FAMILY

    paginate_by = 15

    title = 'Fact-based filtering'


    filterset_class = FactTermValueFilter

    queryset =  InfoObject2Fact.objects.all().exclude(iobject__latest_of=None).prefetch_related('iobject',
                                                                                                'iobject__iobject_type',
                                                                                                'fact__fact_term',
                                                                                                'fact__fact_values').select_related().distinct().order_by('iobject__id')



class InfoObjectView(BasicDetailView):
    """
    Note that below we generate a query set for the facts by hand
    rather than carrying out the queries through the object-query.
    This is because the prefetch_related
    is treated leads to a prefetching of *all* facts, even though
    pagination only displays 100 or 200 or so.
    """

    # Config for Prefetch/SelectRelated Mixins_
    select_related = ()
    prefetch_related = ('iobject_type',
                        'iobject_type__namespace',
                        'identifier__namespace',
    )

    breadcrumbs = (('Dingo',None),
                   ('View',None),
                   ('InfoObject','url.dingos.list.infoobject.generic'),
                   ('[RELOAD]',None)
    )
    model = InfoObject

    max_embedded = 5

    template_name = 'dingos/%s/details/InfoObjectDetails.html' % DINGOS_TEMPLATE_FAMILY

    title = 'Info Object Details'

    @property
    def iobject2facts(self):
        return self.object.fact_thru.all().prefetch_related('fact__fact_term',
                                                             'fact__fact_values',
                                                             'fact__fact_values__fact_data_type',
                                                             'fact__value_iobject_id',
                                                             'fact__value_iobject_id__latest',
                                                             'fact__value_iobject_id__latest__iobject_type',
                                                             'node_id')


    max_embedded = 5

    template_name = 'dingos/%s/details/InfoObjectDetails.html' % DINGOS_TEMPLATE_FAMILY

    title = 'Info Object Details'

    def get_context_data(self, **kwargs):
        context = super(InfoObjectView, self).get_context_data(**kwargs)
        context['max_embedded'] = self.max_embedded

        context['show_NodeID'] = self.request.GET.get('show_nodeid',False)
        context['iobject2facts'] = self.iobject2facts
        try:
            context['highlight'] = self.request.GET['highlight']
        except KeyError:
            context['highlight'] = None

        return context




class InfoObjectJSONView(BasicDetailView):
    # Config for Prefetch/SelectRelated Mixins_
    select_related = ()
    prefetch_related = () # The to_dict function itself defines the necessary prefetch_stuff

    model = InfoObject

    def render_to_response(self, context):
        #return self.get_json_response(json.dumps(context['object'].show_elements(""),indent=2))
        include_node_id = self.request.GET.get('include_node_id',False)

        return self.get_json_response(json.dumps(context['object'].to_dict(include_node_id=include_node_id),indent=2))

    def get_json_response(self, content, **httpresponse_kwargs):
        return http.HttpResponse(content,
                                 content_type='application/json',
                                 **httpresponse_kwargs)




# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import django_filters
from dingos.models import InfoObject, InfoObject2Fact, InfoObjectType, InfoObjectFamily,IdentifierNameSpace
from django.db.models import Count

from django.forms.models import ModelChoiceField

from django.utils.translation import ugettext_lazy as _

from datetime import timedelta
from django.utils.timezone import now

_truncate = lambda dt: dt.replace(hour=0, minute=0, second=0)

class ExtendedDateRangeFilter(django_filters.DateRangeFilter):
    options = {
        '': (_('Any date'), lambda qs, name: qs.all()),
        1: (_('Past 5 minutes'), lambda qs, name: qs.filter(**{
            '%s__gte' % name: now() - timedelta(minutes=5),
            '%s__lt' % name: now() + timedelta(minutes=5),
            })),
        2: (_('Today'), lambda qs, name: qs.filter(**{
            '%s__year' % name: now().year,
            '%s__month' % name: now().month,
            '%s__day' % name: now().day
        })),
        3: (_('Past 24 hours'), lambda qs, name: qs.filter(**{
            '%s__gte' % name: now() - timedelta(days=1),
            '%s__lt' % name: now() + timedelta(days=1),
            })),
        4: (_('Past 7 days'), lambda qs, name: qs.filter(**{
            '%s__gte' % name: _truncate(now() - timedelta(days=7)),
            '%s__lt' % name: _truncate(now() + timedelta(days=1)),
            })),
        5: (_('This month'), lambda qs, name: qs.filter(**{
            '%s__year' % name: now().year,
            '%s__month' % name: now().month
        })),
        6: (_('This year'), lambda qs, name: qs.filter(**{
            '%s__year' % name: now().year,
            })),
        }


class InfoObjectFilter(django_filters.FilterSet):


    iobject_type_qs = InfoObjectType.objects.annotate(num_objects=Count('iobject_set')).\
        filter(num_objects__gt=0).prefetch_related('iobject_family').order_by('iobject_family__name','name')

    iobject_type = django_filters.ModelChoiceFilter(queryset= iobject_type_qs,
                                                    required=None,
                                                    label="InfoObject Type",
                                                    to_field_name='id')

    iobject_type__iobject_family = django_filters.ModelChoiceFilter(
                                                    queryset= InfoObjectFamily.objects.all(),
                                                    required=None,
                                                    label="InfoObject Family",
                                                    to_field_name='id')

    identifier__namespace = django_filters.ModelChoiceFilter(
        queryset= IdentifierNameSpace.objects.all(),
        required=None,
        label="ID Namespace",
        to_field_name='id')

    identifier__uid = django_filters.CharFilter(lookup_type='icontains',
                                                label='ID contains')

    timestamp = ExtendedDateRangeFilter(label="Object Timestamp")

    create_timestamp = ExtendedDateRangeFilter(label="Create/Import Timestamp")

    class Meta:
        model = InfoObject
        fields = ['iobject_type','iobject_type__iobject_family',
                  'identifier__namespace','identifier__uid','timestamp', 'create_timestamp']


class IdSearchFilter(django_filters.FilterSet):

    identifier__uid = django_filters.CharFilter(lookup_type='icontains',
                                                  label='ID contains')



    class Meta:
        model = InfoObject
        fields = ['identifier__namespace','identifier__uid']
        #fields = ['iobject_type','iobject_type__iobject_family']


class FactTermValueFilter(django_filters.FilterSet):

    fact__fact_values__value = django_filters.CharFilter(lookup_type='icontains',
                                                         label='Value contains')

    fact__fact_term__term = django_filters.CharFilter(lookup_type='icontains',
                                                     label='Fact term contains')

    iobject_type_qs = InfoObjectType.objects.annotate(num_objects=Count('iobject_set')). \
        filter(num_objects__gt=0).prefetch_related('iobject_family').order_by('iobject_family__name','name')

    iobject__iobject_type = django_filters.ModelChoiceFilter(queryset= iobject_type_qs,
                                                required=None,
                                                label="InfoObject Type",
                                                to_field_name='id')

    iobject__timestamp = ExtendedDateRangeFilter(label='Object Timestamp')

    iobject__created_timestamp = ExtendedDateRangeFilter(label='Create/Import Timestamp')

    #iobject__iobject_type = django_filters.ModelMultipleChoiceFilter()

    class Meta:
        model = InfoObject2Fact

        fields = ['fact__fact_term__term','fact__fact_values__value','iobject__timestamp','iobject__created_timestamp',
                  'iobject__identifier__namespace','iobject__iobject_type',]







# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import glob
import traceback

from optparse import make_option
from django.core.management.base import BaseCommand, CommandError
from dingos.models import InfoObject

class Command(BaseCommand):
    """
    This class implements the command for importing a generic XML
    file into DINGO.
    """
    args = 'no arguments'
    help = """Regenerates names for all information objects. Use for testing purposes with few objects
              in database rather than for a full database -- otherwise the running time might be rather longish."""

    option_list = BaseCommand.option_list + (
        make_option('-r', '--restrict-to-type-name',
                action='store',
                dest='restrict_by',
                default=None,
                help='String according to which the InfoObjects for which the name is to be generated'
                     'are restricted: only those InfoObjects are touched '
                     'types whose type name contains the provided expression.'),
    )


    def handle(self, *args, **options):
        for io in InfoObject.objects.filter(iobject_type__name__icontains=options['restrict_by']):
            io.set_name()



# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import libxml2


def extract_attributes(element, prefix_key_char='@', dict_constructor=dict):
    """
    Given an XML node, extract a dictionary of attribute key-value pairs.
    Optional arguments:

    - prefix_key_char: if a character is given, the attributes keys
      in the resulting dictionary are prefixed with that character.

    - dict_constructor: the class used to create the resulting dictionary.
      Default is 'dict', but in DINGO, also DingoObjDict may be used.
    """
    result = dict_constructor()
    if element.properties:
        for prop in element.properties:
            if not prop:
                break
            if prop.type == 'attribute':
                try:
                    # First try with namespace. If no namespace exists,
                    # an exception is raised
                    result["%s%s:%s" % (prefix_key_char, prop.ns().name, prop.name)] = prop.content
                except:
                    result["%s%s" % (prefix_key_char, prop.name)] = prop.content
    return result



# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#



from django.utils.http import urlencode

def get_query_string(request, new_params=None, remove=None):
    """
    Given the request, return the query string.

    Parameters can be added or removed as necessary.

    Code snippet taken from Django admin app (views/main.py)

    (c) Copyright Django Software Foundation and individual contributors.


    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of Django nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
    ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    if new_params is None: new_params = {}
    if remove is None: remove = []
    p = dict(request.GET.items())
    for r in remove:
        for k in p.keys():
            if k.startswith(r):
                del p[k]
    for k, v in new_params.items():
        if v is None:
            if k in p:
                del p[k]
        else:
            p[k] = v
    return '?%s' % urlencode(p)



from django.conf.urls import patterns, url
from django.core.urlresolvers import reverse
from django.shortcuts import redirect

from dingos import views
from dingos import filter
from dingos.models import InfoObject

from django_filters.views import FilterView


urlpatterns = patterns('',
    url(r'^View/InfoObject/?$',
        views.InfoObjectList.as_view(),
        name="url.dingos.list.infoobject.generic"),
    url(r'^View/InfoObject/(?P<pk>\d*)/$',
        views.InfoObjectView.as_view(),
        name= "url.dingos.view.infoobject"),
    url(r'^View/InfoObject/(?P<pk>\d*)/json$',
        views.InfoObjectJSONView.as_view(),
        name= "url.dingos.view.infoobject.json"),
    url(r'^Search/SimpleFactSearch$',
        views.SimpleFactSearch.as_view(),
        name="url.dingos.search.fact.simple"),
    url(r'^Search/IdSearch$',
        views.InfoObjectList_Id_filtered.as_view(),
        name="url.dingos.list.infoobject.by_id"),

    # Detail-view with highlight and anchor on certain node
    # solved below with a redirect, because with the 'url' template
    # tag we cannot set an anchor.
    url(r'^View/InfoObject/(?P<pk>\d*)/(?P<node>([A-Z]\d{3,4})?(:[A-Z]\d{3,4})*)/',
        lambda *args, **kwargs: redirect( 
            reverse( 'url.dingos.view.infoobject',
                      kwargs = { 'pk' : int(kwargs['pk']) } ) + '?highlight=%(node)s#%(node)s' % kwargs,
            permanent = True
            ),
        name = 'url.dingos.view.infoobject.redirect2highlight'
    ),

    url(r'^View/InfoOject/Embedded/(?P<pk>\d*)/$', views.InfoObjectsEmbedded.as_view(), name="url.dingos.view.infoobject.embedded"),

                       )



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Deleting field 'BlobStorage._content'
        db.delete_column(u'dingos_blobstorage', 'content')

        # Adding field 'BlobStorage.content'
        db.add_column(u'dingos_blobstorage', 'content',
                      self.gf('django.db.models.fields.TextField')(default='', blank=True),
                      keep_default=False)


    def backwards(self, orm):

        # User chose to not deal with backwards NULL issues for 'BlobStorage._content'
        raise RuntimeError("Cannot reverse this migration. 'BlobStorage._content' and its values cannot be restored.")
        
        # The following code is provided here to aid in writing a correct migration        # Adding field 'BlobStorage._content'
        db.add_column(u'dingos_blobstorage', '_content',
                      self.gf('django.db.models.fields.TextField')(db_column='content'),
                      keep_default=False)

        # Deleting field 'BlobStorage.content'
        db.delete_column(u'dingos_blobstorage', 'content')


    models = {
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'dingos.blobstorage': {
            'Meta': {'object_name': 'BlobStorage'},
            'content': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'sha256': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '64'})
        },
        u'dingos.datatypenamespace': {
            'Meta': {'object_name': 'DataTypeNameSpace'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'blank': 'True'}),
            'uri': ('django.db.models.fields.URLField', [], {'unique': 'True', 'max_length': '255'})
        },
        u'dingos.fact': {
            'Meta': {'object_name': 'Fact'},
            'fact_term': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.FactTerm']"}),
            'fact_values': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['dingos.FactValue']", 'null': 'True', 'symmetrical': 'False'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'value_iobject_id': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'value_of_set'", 'null': 'True', 'to': u"orm['dingos.Identifier']"}),
            'value_iobject_ts': ('django.db.models.fields.DateTimeField', [], {'null': 'True'})
        },
        u'dingos.factdatatype': {
            'Meta': {'unique_together': "(('name', 'name_space'),)", 'object_name': 'FactDataType'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'kind': ('django.db.models.fields.SmallIntegerField', [], {'default': '0'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'name_space': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.DataTypeNameSpace']"})
        },
        u'dingos.factterm': {
            'Meta': {'unique_together': "(('term', 'attribute'),)", 'object_name': 'FactTerm'},
            'attribute': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'term': ('django.db.models.fields.CharField', [], {'max_length': '512'})
        },
        u'dingos.factterm2type': {
            'Meta': {'unique_together': "(('iobject_type', 'fact_term'),)", 'object_name': 'FactTerm2Type'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'fact_data_types': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'fact_term_thru'", 'symmetrical': 'False', 'to': u"orm['dingos.FactDataType']"}),
            'fact_term': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'iobject_type_thru'", 'to': u"orm['dingos.FactTerm']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'iobject_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'fact_term_thru'", 'to': u"orm['dingos.InfoObjectType']"}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '1024', 'blank': 'True'})
        },
        u'dingos.factvalue': {
            'Meta': {'unique_together': "(('value', 'fact_data_type', 'storage_location'),)", 'object_name': 'FactValue'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'fact_data_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.FactDataType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'storage_location': ('django.db.models.fields.SmallIntegerField', [], {'default': '0'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '256', 'blank': 'True'}),
            'value': ('django.db.models.fields.TextField', [], {})
        },
        u'dingos.identifier': {
            'Meta': {'unique_together': "(('uid', 'namespace'),)", 'object_name': 'Identifier'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'latest': ('django.db.models.fields.related.OneToOneField', [], {'related_name': "'latest_of'", 'unique': 'True', 'null': 'True', 'to': u"orm['dingos.InfoObject']"}),
            'namespace': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.IdentifierNameSpace']"}),
            'uid': ('django.db.models.fields.SlugField', [], {'max_length': '255'})
        },
        u'dingos.identifiernamespace': {
            'Meta': {'object_name': 'IdentifierNameSpace'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.SlugField', [], {'max_length': '50', 'blank': 'True'}),
            'uri': ('django.db.models.fields.URLField', [], {'unique': 'True', 'max_length': '255'})
        },
        u'dingos.infoobject': {
            'Meta': {'ordering': "['-timestamp']", 'unique_together': "(('identifier', 'timestamp'),)", 'object_name': 'InfoObject'},
            'create_timestamp': ('django.db.models.fields.DateTimeField', [], {}),
            'facts': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['dingos.Fact']", 'through': u"orm['dingos.InfoObject2Fact']", 'symmetrical': 'False'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'identifier': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.Identifier']"}),
            'iobject_family': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.InfoObjectFamily']"}),
            'iobject_family_revision': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['dingos.Revision']"}),
            'iobject_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.InfoObjectType']"}),
            'iobject_type_revision': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'to': u"orm['dingos.Revision']"}),
            'name': ('django.db.models.fields.CharField', [], {'default': "'Unnamed'", 'max_length': '255', 'blank': 'True'}),
            'timestamp': ('django.db.models.fields.DateTimeField', [], {}),
            'uri': ('django.db.models.fields.URLField', [], {'max_length': '200', 'blank': 'True'})
        },
        u'dingos.infoobject2fact': {
            'Meta': {'ordering': "['node_id__name']", 'object_name': 'InfoObject2Fact'},
            'attributed_fact': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'attributes'", 'null': 'True', 'to': u"orm['dingos.InfoObject2Fact']"}),
            'fact': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'iobject_thru'", 'to': u"orm['dingos.Fact']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'iobject': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'fact_thru'", 'to': u"orm['dingos.InfoObject']"}),
            'node_id': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.NodeID']"})
        },
        u'dingos.infoobjectfamily': {
            'Meta': {'object_name': 'InfoObjectFamily'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.SlugField', [], {'unique': 'True', 'max_length': '256'}),
            'title': ('django.db.models.fields.CharField', [], {'max_length': '1024', 'blank': 'True'})
        },
        u'dingos.infoobjectnaming': {
            'Meta': {'ordering': "['position']", 'object_name': 'InfoObjectNaming'},
            'format_string': ('django.db.models.fields.TextField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'iobject_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.InfoObjectType']"}),
            'position': ('django.db.models.fields.PositiveSmallIntegerField', [], {})
        },
        u'dingos.infoobjecttype': {
            'Meta': {'unique_together': "(('name', 'iobject_family', 'namespace'),)", 'object_name': 'InfoObjectType'},
            'description': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'iobject_family': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.InfoObjectFamily']"}),
            'name': ('django.db.models.fields.SlugField', [], {'max_length': '30'}),
            'namespace': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.DataTypeNameSpace']", 'blank': 'True'})
        },
        u'dingos.marking2x': {
            'Meta': {'object_name': 'Marking2X'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'marking': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'marked_item_thru'", 'to': u"orm['dingos.InfoObject']"}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {})
        },
        u'dingos.nodeid': {
            'Meta': {'object_name': 'NodeID'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'})
        },
        u'dingos.relation': {
            'Meta': {'unique_together': "(('source_id', 'target_id', 'relation_type'),)", 'object_name': 'Relation'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'metadata_id': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'+'", 'null': 'True', 'to': u"orm['dingos.Identifier']"}),
            'relation_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['dingos.Fact']"}),
            'source_id': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'yields_via'", 'null': 'True', 'to': u"orm['dingos.Identifier']"}),
            'target_id': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'yielded_by_via'", 'null': 'True', 'to': u"orm['dingos.Identifier']"})
        },
        u'dingos.revision': {
            'Meta': {'object_name': 'Revision'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '32', 'blank': 'True'})
        }
    }

    complete_apps = ['dingos']


# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#



import logging
import uuid

import libxml2

from collections import deque

from django.utils import timezone

from dingos import *
from dingos.core.datastructures import DingoObjDict
from core.xml_utils import extract_attributes
from dingos.models import dingos_class_map, get_or_create_iobject, Marking2X

import pprint

pp = pprint.PrettyPrinter(indent=2)

logger = logging.getLogger(__name__)



# The constants below are returned by the import function, informing
# about whether an object of the given identifier and timestamp
# already existed.



EXIST_ID_AND_EXACT_TIMESTAMP = "existed"
EXIST_ID_AND_OLDER_TIMESTAMP = "existed_older"
EXIST_ID_AND_NEWER_TIMESTAMP = "existed_newer"
EXIST_PLACEHOLDER = 'exist_placeholder'
NO_EXISTING_OBJECT_FOUND = False


class DingoImportHandling(object):
    def __init__(self, *args, **kwargs):
        logger.debug("Instantiated DingoImportHandling")
        self._DCM = kwargs.get('dingos_class_map', dingos_class_map)
        try:
            del (kwargs['dingos_class_map'])
        except:
            pass


    def get_latest_revision_of_iobject_by_uid(self, namespace_uri, uid):
        """
        Given an information object identifier (represented by uid and uri of the identifier's namespace),
        this function returns the latest revision of an information
        object. If no object of the given identifier exists, None is returned.
        """
        uid_objects = self._DCM['InfoObject'].objects.filter(identifier__namespace__uri=namespace_uri,
                                                             identifier__uid=uid).order_by('-timestamp')

        if uid_objects.count() >= 1:
            return uid_objects[0]
        else:
            return None



    def create_iobject(self,
                       identifier_ns_uri=None,
                       uid=None,
                       timestamp=None,
                       create_timestamp=None,
                       iobject_data=None,
                       config_hooks=None,
                       namespace_dict=None,
                       markings=None,
                       import_older_ts=False,
                       iobject_family_name=DINGOS_IOBJECT_FAMILY_NAME,
                       iobject_family_revision_name=DINGOS_REVISION_NAME,
                       iobject_type_name=DINGOS_PLACEHOLDER_TYPE_NAME,
                       iobject_type_namespace_uri=DINGOS_NAMESPACE_URI,
                       iobject_type_revision_name="",
                       ):
        """
        Create an information object:

        - the unique identifier of the information object is specified
          by identifier_ns_uri and uid

        - the timestamp specifies the revision of the information object

        - iobject data must contain the DingoObjDict with the contents
          of the information object that is to be created

        - config_hooks specifies hooking functions for customizing the
          way the DingoObjDict is transformed into facts -- please
          look at the sample import modules for STIX and OpenIOC to
          get an idea of how this is used

        - A list of markings can be provided. Essentially, a marking is simply
          another information object, to which the created information object
          will receive a link. This can be used, e.g., to keep track of provenance
          information (e.g., which import step led to the creation of that information object.

        - Existing defaults for object family and type can be changed via function parameters.
          If nothing is changed, then a DINGO PLACEHOLDER object is created. This special
          object is used by the importer to create objects for forward references found
          during the import.

        Call the function as
    
        (iobject(s),exists) = create_iobject(...)
    
        The 'exists' flag contains information about whether there were already existing
        revisions.

        """

        # Fill in parameter defaults that would be mutable.

        if not iobject_data:
            iobject_data = DingoObjDict()

        if not config_hooks:
            config_hooks = {}

        existing_iobject = None
        existing_timestamp = None

        exists = False
        overwrite = False

        # Check for existing objects of the given uid
        if uid:
            existing_iobject = self.get_latest_revision_of_iobject_by_uid(identifier_ns_uri, uid)
            if existing_iobject:
                existing_timestamp = existing_iobject.timestamp

        if existing_iobject:

            if existing_iobject.iobject_type.name == DINGOS_PLACEHOLDER_TYPE_NAME \
                and existing_iobject.iobject_family.name == DINGOS_IOBJECT_FAMILY_NAME:

                overwrite = True
                exists = EXIST_PLACEHOLDER



            else:
                # We found a non-Placeholder object; let us see, what the timestamp says.
                existing_timestamps = self._DCM['InfoObject'].objects.filter(identifier__uid=uid).filter(
                    identifier__namespace__uri=identifier_ns_uri).values_list('timestamp', flat=True)
                if timestamp in existing_timestamps:
                    exists = EXIST_ID_AND_EXACT_TIMESTAMP
                elif timestamp < existing_timestamp:
                    exists = EXIST_ID_AND_NEWER_TIMESTAMP
                else:
                    exists = EXIST_ID_AND_OLDER_TIMESTAMP

        # If we have found an object of the given id with the exact timestamp or
        # a newer object and the parameter 'import_older_ts' is set to False,
        # we return the latest revision (i.e., object with newest time stamp) of
        # for the given identifier.


        logger.debug("EXISTS flag for %s:%s is: %s" % (identifier_ns_uri, uid, exists))

        if exists==EXIST_ID_AND_EXACT_TIMESTAMP or (not import_older_ts and exists == EXIST_ID_AND_NEWER_TIMESTAMP):
            return (existing_iobject, exists)

        else:
            # Otherwise, we create the object

            if not uid:
                uid = uuid.uuid1()
            logger.info("Creating %s:%s with timestamp %s" % (identifier_ns_uri,
                                                              uid,
                                                              '{:%d-%m-%Y:%H:%M:%S}.{:03d}'.format(timestamp, timestamp.microsecond // 1000)))
            iobject, created = get_or_create_iobject(uid,
                                                     identifier_ns_uri,
                                                     iobject_type_name,
                                                     iobject_type_namespace_uri,
                                                     iobject_type_revision_name,
                                                     iobject_family_name,
                                                     iobject_family_revision_name,
                                                     identifier_namespace_name=None,
                                                     timestamp=timestamp,
                                                     create_timestamp=create_timestamp,
                                                     overwrite=existing_iobject,
                                                     dingos_class_map=self._DCM,
                                                    )

            # After creating the object, we write the facts to the object.
            # We overwrite in the special case that a PLACEHOLDER was found.

            if created or overwrite:
                iobject.from_dict(iobject_data, config_hooks=config_hooks, namespace_dict=namespace_dict)


            # We adjust the back pointer in the identifier table to the latest version
            # of the object
            if not exists or exists == EXIST_ID_AND_OLDER_TIMESTAMP or exists == EXIST_PLACEHOLDER:
                iobject.identifier.latest = iobject
                iobject.identifier.save()

            if markings:
                # If markings were given, we create the marking.

                for marking in markings:
                    Marking2X.objects.create(marked=iobject,
                                             marking=marking)

            return (iobject, exists)


    def create_marking_iobject(self,
                               uid=None,
                               timestamp=timezone.now(),
                               metadata_dict=None,
                               id_namespace_uri=DINGOS_DEFAULT_ID_NAMESPACE_URI,
                               iobject_family_name=DINGOS_IOBJECT_FAMILY_NAME,
                               iobject_family_revison_name=DINGOS_REVISION_NAME,
                               iobject_type_name=DINGOS_DEFAULT_IMPORT_MARKING_TYPE_NAME,
                               iobject_type_namespace_uri=DINGOS_NAMESPACE_URI,
                               iobject_type_revision_name=DINGOS_REVISION_NAME,
                               ):
        """
        A specialized version of create_iobject with defaults set such that a default marking object is created.
        """
        if not uid:
            uid = uuid.uuid1()

        iobject, created = self.create_iobject(iobject_family_name=iobject_family_name,
                                                      iobject_family_revision_name=iobject_family_revison_name,
                                                      iobject_type_name=iobject_type_name,
                                                      iobject_type_namespace_uri=iobject_type_namespace_uri,
                                                      iobject_type_revision_name=iobject_type_revision_name,
                                                      iobject_data=metadata_dict,
                                                      uid=uid,
                                                      identifier_ns_uri=id_namespace_uri,
                                                      timestamp=timestamp,
                                                      )

        return iobject

    def xml_import(self,
                   xml_fname=None,
                   xml_content=None,
                   ns_mapping=None,
                   embedded_predicate=None,
                   id_and_revision_extractor=None,
                   extract_empty_embedded=False,
                   keep_attrs_in_created_reference=True,
                   transformer=None):
        """
        This is the generic XML import function for dingos. Its parameters
        are as follows:

        - xml_fname: Filename of the XML file to be read
        - xml_content: Alternatively, the xml_content can be provided as string
          or as XMLNode (i.e., a piece of XML that has already been parsed)
        - ns_mapping: A dictionary that may already contain mappings of namespaces
          to namespace URIs. Attention: this dictionary will be enriched with namespace
          information found in the XML file!!!
        - embedded_predicate:
          A function that, when given an XML node and a child node, determines whether
          the child node should be treated as separate entity that has been embedded.
          Please refer to existing import modules such as for STIX or OpenIOC for
          examples of how to use this parameter.

        - id_and_revision_extractor:
          A function that, when given an XML node and a child node, determines whether
          this node specifies an identifier and possibly a timestamp.
          Please refer to existing import modules such as for STIX or OpenIOC for
          examples of how to use this parameter.

        - extract_empty_embedded:
          A flag (True/False) governing whether elements that are recognized as
          being embedded but contain no childs should be extracted as separate
          object or not. The default is False; the setting "True" may be necessary
          in cases where there are embedded objects that contain all its information
          in attributes rather than using child elements.

        - keep_attrs_in_created_reference:
          A flag (True/False) governing the shape of the reference created for
          an embedded object: when an embedding is recognized, it is extracted
          and a reference using 'idref' inserted instead. If 'keep_attrs_in_created_reference'
          is True, then the top-level attributes contained in the found embedded object
          are also retained in the reference.


        - transformer:
          A function that, when given an element name and a DingoObjDict containing
          the result of importing the contents under the element of the given name,
          may or may not change the element name and transform the DingoObjDict.
          Please refer to existing import MANTIS modules such as for OpenIOC for
          examples of how to use this parameter.


        Note: a good starting point for understanding how to use the python bindings
        of libxml2 is http://mikekneller.com/kb/python/libxml2python/part1.
        """

        generated_id_count = {}



        # Fill defaults
        if not ns_mapping:
            nas_mapping = {}
        if not transformer:
            transformer = lambda x, y: (x, y)

        # We use the _import_pending_stack to hold extracted embedded objects
        # that still need to be processed
        _import_pending_stack = deque()

        # We collect the read embedded objects in the following list
        embedded_objects = deque()

        def xml_import_(element, depth=0, type_info=None, inherited_id_and_rev_info=None):
            """
            Recursive import function
            """

            if not inherited_id_and_rev_info:
                inherited_id_and_rev_info = main_id_and_rev_info.copy()

            fresh_inherited_id_and_rev_info = inherited_id_and_rev_info.copy()


            if element.name == 'comment':
                return None


            #try:
            #    namespace = element.ns()
            #    ns_mapping[namespace.name]=namespace.content
            #except:
            #    pass

            result = DingoObjDict()

            # Add properties to result dictionary for this element

            if element.properties:
                for prop in element.properties:

                    if not prop:
                        break
                    if prop.type == 'attribute':
                        try:
                            result["@%s:%s" % (prop.ns().name, prop.name)] = prop.content
                        except:
                            result["@%s" % prop.name] = prop.content



            # see if there is a namespace

            try:
                ns = element.ns().name
                result["@@ns"] = ns
            except:
                pass

            # prepare list for keeping resulting dictionaries of child elements

            element_dicts = []

            # While looking at the child-elements, we have to keep track
            # of certain data.

            # Firstly: keep track whether we have seen text content that is not whitespace --
            # if that happens, we know that this element contains mixed
            # content and we will back off and just dump the element contents
            # as is as value into the dictionary.

            non_ws_content = False

            # Secondly: If the element contains cdata stuff, we will see
            # that one (well, the only) child has type cdata. So if
            # we find such a child, we set the flag

            cdata_content = False

            # Thirdly: we keep track of how many different child-element-names
            # we have seen. 
            #
            # - If we find that we have exactly one distinct name,
            #   we will generate a dictionary of form
            #             {<Element_Name> : [ <list of child elemen dictionaries> ]}
            # - If we find that we have as many distinct names as we
            #   child elements, we create a dictionary mapping each child element
            #   name to its dictionary representation
            # - If we find that we have less child element names than
            #   we have children, we know that at least one name
            #   occured more than once. Our dictionary representation cannot
            #   deal with that, and we back off and dump the contents as they
            #   are marked as 'xml' content with the '@@type' attribute.

            name_set = {}

            previous_seen_child = None
            double_occurrance = False

            element_child_count = 0
            child = element.children
            while child is not None:

                #if child_name=='comment':
                #    pass
                if child.name == 'text':
                    # If we have non-whitespace content in one of the children,
                    # we set the non_ws_content flag
                    content = child.content.strip()
                    if content != "":
                        non_ws_content = True

                elif child.type == 'cdata':
                    logger.debug("!!!!FOUND CDATA")
                    # If one of the children (actually, it should be the only child)
                    # has type cdata, we know that the parent element contains cdata
                    # and set the cdata_content flag accordingly
                    cdata_content = True
                else:
                    # we have found an element, so we recurse into it.
                    element_child_count += 1
                    if previous_seen_child and (child.name in name_set) and (not child.name == previous_seen_child):
                        double_occurrance = True

                    name_set[child.name] = None

                    if embedded_predicate:
                        embedded_ns = embedded_predicate(element, child, ns_mapping)
                        logger.debug("Embedded ns is %s" % embedded_ns)

                        if embedded_ns:
                            inherited_id_and_rev_info = fresh_inherited_id_and_rev_info.copy()
                            # There is an embedded object. We therefore
                            # replace the contents of the element with an element
                            # containing an idref (and, since we might need them,
                            # all attributes of the embedded element)

                            if type(embedded_ns) == type({}):
                                # If necessary, the embedded_predicate can return more information
                                # than namespace information, namely we can can hand down
                                # id and revision info that has been derived wenn the embedding
                                # was detected. For backward compatibility,
                                # we further allow returning of a string; if, however,
                                # a dictionary is returned, there is id_and_revision_info.
                                id_and_revision_info = embedded_ns.get('id_and_revision_info',
                                                                       id_and_revision_extractor(child))
                                embedded_ns = embedded_ns.get('embedded_ns',None)
                            else:
                                id_and_revision_info = id_and_revision_extractor(child)


                            # See whether stuff needs to be inherited
                            if not 'id' in id_and_revision_info or not id_and_revision_info['id']:
                                if 'id' in inherited_id_and_rev_info:
                                    parent_id = inherited_id_and_rev_info['id']
                                    if parent_id in generated_id_count:
                                        gen_counter = generated_id_count[parent_id]
                                        gen_counter +=1
                                    else:
                                        gen_counter = 0
                                    generated_id_count[parent_id] = gen_counter
                                    (parent_namespace, parent_uid) = parent_id.split(':')
                                    generated_id = "%s:emb%s-in-%s" % (parent_namespace,gen_counter,parent_uid)
                                    logger.info("Found embedded %s without id and generated id %s" % (element.name,generated_id))
                                    id_and_revision_info['id'] = generated_id
                                    id_and_revision_info['id_inherited'] = True
                                else:
                                    logger.error("Attempt to import object (element name %s) without id -- object is ignored" % elt_name)

                                    #cybox_id = gen_cybox_id(iobject_type_name)

                            if not id_and_revision_info.get('timestamp', None):
                                if inherited_id_and_rev_info and 'timestamp' in inherited_id_and_rev_info:
                                    id_and_revision_info['timestamp'] = inherited_id_and_rev_info['timestamp']
                                    id_and_revision_info['ts_inherited'] = True
                            else:
                                inherited_id_and_rev_info['timestamp'] = id_and_revision_info['timestamp']


                            if 'id' in id_and_revision_info:
                                # If the identifier has no namespace info (this may occur, e.g. for
                                # embedded OpenIOC in STIX, we take the namespace inherited from  the
                                # embedding object
                                if (not ':' in id_and_revision_info['id']
                                    and inherited_id_and_rev_info['id']
                                    and ':' in inherited_id_and_rev_info['id']):
                                    id_and_revision_info['id'] = "%s:%s" % (inherited_id_and_rev_info['id'].split(':')[0],
                                                                            id_and_revision_info['id'])
                                    id_and_revision_info['ns_inherited'] = True

                                inherited_id_and_rev_info['id'] = id_and_revision_info['id']



                            if keep_attrs_in_created_reference:
                                reference_dict = extract_attributes(child, prefix_key_char='@',
                                                                    dict_constructor=DingoObjDict)
                            else:
                                reference_dict = DingoObjDict()

                            reference_dict['@idref'] = id_and_revision_info['id']

                            reference_dict['@@timestamp'] = id_and_revision_info['timestamp']

                            try:
                                reference_dict['@@ns'] = child.ns().name
                            except:
                                reference_dict['@@ns'] = None
                            if embedded_ns == True:
                                embedded_ns = None
                            logger.debug("Setting embedded type info to %s" % embedded_ns)
                            reference_dict['@@embedded_type_info'] = embedded_ns

                            element_dicts.append((child.name, reference_dict))
                            if (child.children or child.content) \
                                or extract_empty_embedded \
                                or 'extract_empty_embedded' in id_and_revision_info:

                                id_and_revision_info['inherited'] = fresh_inherited_id_and_rev_info.copy()
                                if 'inherited' in id_and_revision_info['inherited']:
                                    for key in id_and_revision_info['inherited']['inherited']:
                                        if not key in id_and_revision_info['inherited']:
                                            id_and_revision_info['inherited'][key] = id_and_revision_info['inherited']['inherited'][key]
                                    del(id_and_revision_info['inherited']['inherited'])

                                logger.debug(
                                    "Adding XML subtree starting with element %s and type info %s to pending stack." % (
                                    id_and_revision_info, embedded_ns))
                                _import_pending_stack.append((id_and_revision_info, embedded_ns, child))
                            else:
                                # For example, in cybox 1.0, the following occurs::
                                #         <EmailMessageObj:File xsi:type="FileObj:FileObjectType" object_reference="cybox:object-3cf6a958-5c3f-11e2-a06c-0050569761d3"/>
                                # This is only a reference and may not be confused with the definition of the object,
                                # which occurs someplace else -- otherwise, the (almost) empty reference is created as object
                                # and may overwrite the object resulting from the real definition.
                                logger.info(
                                    "Not adding element %s with type info %s to pending stack because element is empty." % (
                                    id_and_revision_info, embedded_ns))
                        else:
                            child_import = xml_import_(child, depth + 1, inherited_id_and_rev_info=inherited_id_and_rev_info)
                            if child_import:
                                element_dicts.append(child_import)
                    else:
                        child_import = xml_import_(child, depth + 1, inherited_id_and_rev_info=inherited_id_and_rev_info)
                        if child_import:
                            element_dicts.append(child_import)

                child = child.next
                # now, we decide what to do with this node

            distinct_child_count = len(name_set.keys())

            if distinct_child_count == 0:
                # No child elements were detected, so we dump the content into
                # the value
                result['_value'] = element.content
                if cdata_content:
                    # If this is a cdata element, we mark it as such
                    result['@@content_type'] = 'cdata'
            elif non_ws_content == True:
                # We have mixed content, so we dump it
                sub_child = element.children
                serialization = ''
                while sub_child:
                    serialization += sub_child.serialize()
                    sub_child = sub_child.next
                result['_value'] = serialization.strip()
                #result['_value']=element.serialize()
                result['@@content_type'] = 'mixed'
            elif double_occurrance: # distinct_child_count >1 and (distinct_child_count) < element_child_count:
                # We have a structure our dictionary representation cannot
                # deal with -- so we dump it

                logger.warning("Cannot deal with XML structure of %s (children %s, count %s): will dump to value" % (
                element.name, name_set.keys(), element_child_count))
                sub_child = element.children
                serialization = ''
                while sub_child:
                    serialization += sub_child.serialize()
                    sub_child = sub_child.next
                result['_value'] = serialization.strip()
                #result['_value']=element.serialize()
                result['@@content_type'] = 'xml'

            else:

                previously_written_name = None
                for (name, element_dict) in element_dicts:

                    if not previously_written_name or name != previously_written_name:
                        result[name] = element_dict
                        previously_written_name = name
                    else: # if name == previously_written_name:
                        if type(result[name]) == type([]):
                            result[name].append(element_dict)
                        else:
                            result[name] = [result[name], element_dict]
            if type_info:
                result['@@embedded_type_info'] = type_info

            element_ns = None
            try:
                element_ns = element.ns().name
            except:
                pass

            return transformer(element.name, result)


        if xml_content:

            if isinstance(xml_content,libxml2.xmlNode):
                root = xml_content
            else:
                doc = libxml2.parseDoc(xml_content)
                root = doc.getRootElement()

        else:
            doc = libxml2.recoverFile(xml_fname)
            root = doc.getRootElement()
            with open(xml_fname, 'r') as content_file:
                xml_content = content_file.read()







        # Extract namespace information (if any)
        try:
            ns_def = root.nsDefs()
            while ns_def:
                ns_mapping[ns_def.name] = ns_def.content
                ns_def = ns_def.next
        except:
            pass

        # Extract ID and timestamp for root element

        main_id_and_rev_info = id_and_revision_extractor(root)

        # Call the internal recursive function. This returns
        # - name of the top-level element
        # - DingoObjDict resulting from import
        # As side effect, it pushes the XML nodes of
        # found embedded objects onto the pending stack

        (main_elt_name, main_elt_dict) = xml_import_(root, 0)

        # We now go through the pending stack.
        # For each found embedded object, xml_import_ pushes
        # the following triple on the stack:
        # - id_and_revision_info: A dictionary, containing
        #   identifier and (possibly) timestamp information
        #   for that object
        # - type_info: Information about the type of the
        #   embedded object (can be None)
        # - the XML node that describes the embedded object

        do_not_process_list = []

        while _import_pending_stack:
            (id_and_revision_info, type_info, elt) = _import_pending_stack.pop()
            if 'defer_processing' in id_and_revision_info:
                do_not_process_list.append((id_and_revision_info,type_info,elt))

            else:
                (elt_name, elt_dict) = xml_import_(elt, 0,
                                                   type_info=type_info,
                                                   inherited_id_and_rev_info=id_and_revision_info.copy())

                embedded_objects.append({'id_and_rev_info': id_and_revision_info,
                                             'elt_name': elt_name,
                                             'dict_repr': elt_dict})

        result= {'id_and_rev_info': main_id_and_rev_info,
                'elt_name': main_elt_name,
                'dict_repr': main_elt_dict,
                'embedded_objects': embedded_objects,
                'unprocessed' : do_not_process_list,
                'file_content': xml_content}


        #pp.pprint(result)

        return result




    
    



# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import pprint

pp = pprint.PrettyPrinter(indent=2)

def print_arguments():
    """
    Decorator that provides the wrapped function with an attribute 'print_arguments'
    containing just those keyword arguments actually passed in to the function.

    To use the decorator for debugging, preface the function into whose calls
    you are interested with '@print_arguments()'
    """
    def decorator(function):
        def inner(*args, **kwargs):
            if args:
                print "Passed arguments:"
                for i in args:
                    pp.pprint(i)
            print "Passed keyword arguments:"
            pp.pprint(kwargs)
            return function(*args, **kwargs)
        return inner
    return decorator


# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


def listify(x):
    """
    listify(x) returns a singleton list containing x if x is not already a list.
    If x is a list, x itself is returned.

    This function is used when traversing dictionaries created by the DINGO XML
    import: because the XML import knows nothing about the XML schema, it
    cannot know whether a given single element is actually a list of length
    one. For example::

        <foo>
           <bar>1</bar>
        </foo>

    would lead to::

        {'foo' : {'bar' : 1 }}

    whereas::

        <foo>
           <bar>1</bar>
           <bar>2</bar>
        </foo>

    would lead to::

        {'foo' : {'bar' : [1,2]}}

    So when traversing a dictionary structure we might expect a list
    but in case of singletons there will be no list. 'listify' can
    be used instead of case-distinctions.
    """
    if type(x) == type([]):
        return x
    else:
        return [x]


def set_dict(dictionary, value, command='set', *keys):
    """
    set_dict takes a dictionary, the value to
    enter into the dictionary, a command of what
    to do with the value, and a sequence of keys.

    d = {}

    set_dict(d,1,'append','level 1','level 2')

    -> d['level 1']['level 2'] = [1]

    set_dict(d,2,'append','level 1','level 2')

    -> d['level 1']['level 2'] = [1,2]

    """

    existing = dictionary
    for i in range(0, len(keys) - 1):
        if keys[i] in existing:
            existing = existing[keys[i]]
        else:
            existing[keys[i]] = existing.__class__()
            existing = existing[keys[i]]
    if command == 'set':
        existing[keys[len(keys) - 1]] = value
    elif command == 'append':
        if keys[len(keys) - 1] in existing:
            existing[keys[len(keys) - 1]].append(value)
        else:
            existing[keys[len(keys) - 1]] = [value]
    elif command == 'set_or_append':
        if keys[len(keys) - 1] in existing:
            if type(keys[len(keys) - 1]) == type([]):
                existing[keys[len(keys) - 1]].append(value)
            else:
                existing[keys[len(keys) - 1]] = [existing[keys[len(keys) - 1]], value]
        else:
            existing[keys[len(keys) - 1]] = value
    elif command == 'insert':
        if keys[len(keys) - 1] in existing:
            if not value in existing[keys[len(keys) - 1]]:
                existing[keys[len(keys) - 1]].append(value)
        else:
            existing[keys[len(keys) - 1]] = [value]


def get_dict(dictionary, *keys):
    """
    This function allows traversals over several keys to
    be performed by passing a list of keys::

    get_dict(d,key1,key2,key3) = d[key1][key2][key3]

    """

    existing = dictionary
    for i in range(0, len(keys)):
        if keys[i] in existing:
            existing = existing[keys[i]]
        else:
            return None
    return existing


def search_by_re_list(re_list, text):
    """
    Given a list of compiled regular expressions,
    try to search in text with each matcher until the first
    match occurs. Return the group-dict for that first match.
    """
    for matcher in re_list:
        m = matcher.search(text)
        if m:
            return m.groupdict()
    return None



# Copyright (c) Siemens AG, 2013
#
# This file is part of MANTIS.  MANTIS is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2
# of the License, or(at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import glob
import traceback

from optparse import make_option
from django.core.management.base import CommandError
from dingos.importer import Generic_XML_Import, DingoImportCommand



class Command(DingoImportCommand):
    """
    This class implements the command for importing a generic XML
    file into DINGO.
    """
    args = 'path/to/xml-file path/to/xml-file ... (you can use wildcards)'
    help = 'Imports xml files of specified paths into DINGO'

    option_list = DingoImportCommand.option_list + (
        make_option('-i', '--identifier',
                    action='store',
                    dest='uid',
                    default=None,
                    help="""Unique identifier for imported information object (optional: otherwise the SHA256
                            of the file contents is used. Only possible if exactly one file is specified."""),
    )

    # def handle(self, *args, **options):
    #
    #     # The function create_import_marking inherited from
    #     # DingoImport command is able to create a dictionary
    #     # structure for a marking with object resulting
    #     # from the import command will be marked.
    #
    #     marking = self.create_import_marking(args,options)
    #
    #     if marking:
    #         markings = [marking]
    #     else:
    #         markings = []
    #
    #     if len(args) > 1 and options['identifer']:
    #         raise CommandError('Option --identifier not supported for more than one file per import.')
    #
    #     for arg in args:
    #         print "Starting processing"
    #         for filename in glob.glob(arg):
    #             try:
    #                 print "Starting import of %s" % filename
    #                 GenericImporter = Generic_XML_Import()
    #                 GenericImporter.xml_import(filename,markings,uid=options['identifier'])
    #             except Exception, err:
    #                 print traceback.format_exc()
    #                 print Exception, err
    #                 raise CommandError('Error %s occurred for %s' % (err,filename))
